package com.sun.imageio.plugins.gif;

import java.util.Iterator;
import javax.imageio.IIOException;
import java.awt.image.DataBufferByte;
import sun.awt.image.ByteComponentRaster;
import java.awt.image.ComponentSampleModel;
import java.awt.image.Raster;
import com.sun.imageio.plugins.common.LZWCompressor;
import org.w3c.dom.NodeList;
import java.util.Arrays;
import javax.imageio.metadata.IIOMetadataNode;
import com.sun.imageio.plugins.common.PaletteBuilder;
import java.awt.image.RenderedImage;
import javax.imageio.IIOImage;
import java.nio.ByteOrder;
import javax.imageio.ImageTypeSpecifier;
import javax.imageio.metadata.IIOInvalidTreeException;
import javax.imageio.metadata.IIOMetadata;
import java.awt.image.IndexColorModel;
import java.awt.image.SampleModel;
import java.awt.image.ColorModel;
import javax.imageio.ImageWriteParam;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.io.IOException;
import javax.imageio.stream.ImageOutputStream;
import javax.imageio.ImageWriter;

public class GIFImageWriter extends ImageWriter
{
    private static final boolean DEBUG = false;
    static final String STANDARD_METADATA_NAME = "javax_imageio_1.0";
    static final String STREAM_METADATA_NAME = "javax_imageio_gif_stream_1.0";
    static final String IMAGE_METADATA_NAME = "javax_imageio_gif_image_1.0";
    private ImageOutputStream stream;
    private boolean isWritingSequence;
    private boolean wroteSequenceHeader;
    private GIFWritableStreamMetadata theStreamMetadata;
    private int imageIndex;
    
    private static int getNumBits(int i) throws IOException {
        int n;
        switch (i) {
            case 2: {
                n = 1;
                break;
            }
            case 4: {
                n = 2;
                break;
            }
            case 8: {
                n = 3;
                break;
            }
            case 16: {
                n = 4;
                break;
            }
            case 32: {
                n = 5;
                break;
            }
            case 64: {
                n = 6;
                break;
            }
            case 128: {
                n = 7;
                break;
            }
            case 256: {
                n = 8;
                break;
            }
            default: {
                throw new IOException("Bad palette length: " + i + "!");
            }
        }
        return n;
    }
    
    private static void computeRegions(Rectangle r, Dimension dimension, ImageWriteParam imageWriteParam) {
        int sourceXSubsampling = 1;
        int sourceYSubsampling = 1;
        if (imageWriteParam != null) {
            int[] sourceBands = imageWriteParam.getSourceBands();
            if (sourceBands != null && (sourceBands.length != 1 || sourceBands[0] != 0)) {
                throw new IllegalArgumentException("Cannot sub-band image!");
            }
            Rectangle sourceRegion = imageWriteParam.getSourceRegion();
            if (sourceRegion != null) {
                r.setBounds(sourceRegion.intersection(r));
            }
            int subsamplingXOffset = imageWriteParam.getSubsamplingXOffset();
            int subsamplingYOffset = imageWriteParam.getSubsamplingYOffset();
            r.x += subsamplingXOffset;
            r.y += subsamplingYOffset;
            r.width -= subsamplingXOffset;
            r.height -= subsamplingYOffset;
            sourceXSubsampling = imageWriteParam.getSourceXSubsampling();
            sourceYSubsampling = imageWriteParam.getSourceYSubsampling();
        }
        dimension.setSize((r.width + sourceXSubsampling - 1) / sourceXSubsampling, (r.height + sourceYSubsampling - 1) / sourceYSubsampling);
        if (dimension.width <= 0 || dimension.height <= 0) {
            throw new IllegalArgumentException("Empty source region!");
        }
        return;
    }
    
    private static byte[] createColorTable(ColorModel colorModel, SampleModel sampleModel) {
        byte[] array;
        if ((colorModel instanceof IndexColorModel)) {
            IndexColorModel indexColorModel = (IndexColorModel)colorModel;
            int j = indexColorModel.getMapSize();
            int gifPaletteSize = getGifPaletteSize(j);
            byte[] r = new byte[gifPaletteSize];
            byte[] g = new byte[gifPaletteSize];
            byte[] b = new byte[gifPaletteSize];
            indexColorModel.getReds(r);
            indexColorModel.getGreens(g);
            indexColorModel.getBlues(b);
            for (; j < gifPaletteSize; ++j) {
                r[j] = r[0];
                g[j] = g[0];
                b[j] = b[0];
            }
            array = new byte[3 * gifPaletteSize];
            int n3 = 0;
            for (int k = 0; k < gifPaletteSize; ++k) {
                array[n3++] = r[k];
                array[n3++] = g[k];
                array[n3++] = b[k];
            }
        }
        else if (sampleModel.getNumBands() == 1) {
            int n = sampleModel.getSampleSize()[0];
            if (n > 8) {
                n = 8;
            }
            int n2 = 3 * (1 << n);
            array = new byte[n2];
            for (int i = 0; i < n2; ++i) {
                array[i] = (byte)(i / 3);
            }
        }
        else {
            array = null;
        }
        return array;
    }
    
    private static int getGifPaletteSize(int n) {
        if (n <= 2) {
            return 2;
        }
        --n;
        n |= n >> 1;
        n |= n >> 2;
        n |= n >> 4;
        n |= n >> 8;
        n |= n >> 16;
        return n + 1;
    }
    
    public GIFImageWriter(GIFImageWriterSpi originatingProvider) {
        super(originatingProvider);
        this.stream = null;
        this.isWritingSequence = false;
        this.wroteSequenceHeader = false;
        this.theStreamMetadata = null;
        this.imageIndex = 0;
    }
    
    @Override
    public boolean canWriteSequence() {
        return true;
    }
    
    private void convertMetadata(String s, IIOMetadata iioMetadata, IIOMetadata iioMetadata2) {
        String s2 = null;
        String nativeMetadataFormatName = iioMetadata.getNativeMetadataFormatName();
        if (nativeMetadataFormatName != null && nativeMetadataFormatName.equals(s)) {
            s2 = s;
        }
        else {
            String[] extraMetadataFormatNames = iioMetadata.getExtraMetadataFormatNames();
            if (extraMetadataFormatNames != null) {
                int i = 0;
                while (i < extraMetadataFormatNames.length) {
                    if (extraMetadataFormatNames[i].equals(s)) {
                        s2 = s;
                        break;
                    }
                    ++i;
                }
            }
        }
        if (s2 == null && iioMetadata.isStandardMetadataFormatSupported()) {
            s2 = "javax_imageio_1.0";
        }
        if (s2 != null) {
            try {
                iioMetadata2.mergeTree(s2, iioMetadata.getAsTree(s2));
            }
            catch (IIOInvalidTreeException ex) {}
        }
    }
    
    @Override
    public IIOMetadata convertStreamMetadata(IIOMetadata iioMetadata, ImageWriteParam imageWriteParam) {
        if (iioMetadata == null) {
            throw new IllegalArgumentException("inData == null!");
        }
        IIOMetadata defaultStreamMetadata = this.getDefaultStreamMetadata(imageWriteParam);
        this.convertMetadata("javax_imageio_gif_stream_1.0", iioMetadata, defaultStreamMetadata);
        return defaultStreamMetadata;
    }
    
    @Override
    public IIOMetadata convertImageMetadata(IIOMetadata iioMetadata, ImageTypeSpecifier imageTypeSpecifier, ImageWriteParam imageWriteParam) {
        if (iioMetadata == null) {
            throw new IllegalArgumentException("inData == null!");
        }
        if (imageTypeSpecifier == null) {
            throw new IllegalArgumentException("imageType == null!");
        }
        GIFWritableImageMetadata gifWritableImageMetadata = (GIFWritableImageMetadata)this.getDefaultImageMetadata(imageTypeSpecifier, imageWriteParam);
        int interlaceFlag = gifWritableImageMetadata.interlaceFlag;
        this.convertMetadata("javax_imageio_gif_image_1.0", iioMetadata, gifWritableImageMetadata);
        if (imageWriteParam != null && imageWriteParam.canWriteProgressive() && imageWriteParam.getProgressiveMode() != 3) {
            gifWritableImageMetadata.interlaceFlag = interlaceFlag;
            return gifWritableImageMetadata;
        }
        return gifWritableImageMetadata;
    }
    
    @Override
    public void endWriteSequence() throws IOException {
        if (this.stream == null) {
            throw new IllegalStateException("output == null!");
        }
        if (!this.isWritingSequence) {
            throw new IllegalStateException("prepareWriteSequence() was not invoked!");
        }
        this.writeTrailer();
        this.resetLocal();
        return;
    }
    
    @Override
    public IIOMetadata getDefaultImageMetadata(ImageTypeSpecifier imageTypeSpecifier, ImageWriteParam imageWriteParam) {
        GIFWritableImageMetadata gifWritableImageMetadata = new GIFWritableImageMetadata();
        SampleModel sampleModel = imageTypeSpecifier.getSampleModel();
        Rectangle rectangle = new Rectangle(sampleModel.getWidth(), sampleModel.getHeight());
        Dimension dimension = new Dimension();
        computeRegions(rectangle, dimension, imageWriteParam);
        gifWritableImageMetadata.imageWidth = dimension.width;
        gifWritableImageMetadata.imageHeight = dimension.height;
        if (imageWriteParam != null && imageWriteParam.canWriteProgressive() && imageWriteParam.getProgressiveMode() == 0) {
            gifWritableImageMetadata.interlaceFlag = false;
        }
        else {
            gifWritableImageMetadata.interlaceFlag = true;
        }
        ColorModel colorModel = imageTypeSpecifier.getColorModel();
        gifWritableImageMetadata.localColorTable = createColorTable(colorModel, sampleModel);
        if ((colorModel instanceof IndexColorModel)) {
            int transparentPixel = ((IndexColorModel)colorModel).getTransparentPixel();
            if (transparentPixel != -1) {
                gifWritableImageMetadata.transparentColorFlag = true;
                gifWritableImageMetadata.transparentColorIndex = transparentPixel;
            }
        }
        return gifWritableImageMetadata;
    }
    
    @Override
    public IIOMetadata getDefaultStreamMetadata(ImageWriteParam imageWriteParam) {
        GIFWritableStreamMetadata gifWritableStreamMetadata = new GIFWritableStreamMetadata();
        gifWritableStreamMetadata.version = "89a";
        return gifWritableStreamMetadata;
    }
    
    @Override
    public ImageWriteParam getDefaultWriteParam() {
        return new GIFImageWriteParam(this.getLocale());
    }
    
    @Override
    public void prepareWriteSequence(IIOMetadata iioMetadata) throws IOException {
        if (this.stream == null) {
            throw new IllegalStateException("Output is not set.");
        }
        this.resetLocal();
        if (iioMetadata == null) {
            this.theStreamMetadata = (GIFWritableStreamMetadata)this.getDefaultStreamMetadata(null);
        }
        else {
            this.convertMetadata("javax_imageio_gif_stream_1.0", iioMetadata, this.theStreamMetadata = new GIFWritableStreamMetadata());
        }
        this.isWritingSequence = true;
        return;
    }
    
    @Override
    public void reset() {
        super.reset();
        this.resetLocal();
    }
    
    private void resetLocal() {
        this.isWritingSequence = false;
        this.wroteSequenceHeader = false;
        this.theStreamMetadata = null;
        this.imageIndex = 0;
    }
    
    @Override
    public void setOutput(Object output) {
        super.setOutput(output);
        if (output != null) {
            if (!(output instanceof ImageOutputStream)) {
                throw new IllegalArgumentException("output is not an ImageOutputStream");
            }
            (this.stream = (ImageOutputStream)output).setByteOrder(ByteOrder.LITTLE_ENDIAN);
        }
        else {
            this.stream = null;
        }
    }
    
    @Override
    public void write(IIOMetadata iioMetadata, IIOImage iioImage, ImageWriteParam imageWriteParam) throws IOException {
        if (this.stream == null) {
            throw new IllegalStateException("output == null!");
        }
        if (iioImage == null) {
            throw new IllegalArgumentException("iioimage == null!");
        }
        if (iioImage.hasRaster()) {
            throw new UnsupportedOperationException("canWriteRasters() == false!");
        }
        this.resetLocal();
        GIFWritableStreamMetadata gifWritableStreamMetadata;
        if (iioMetadata == null) {
            gifWritableStreamMetadata = (GIFWritableStreamMetadata)this.getDefaultStreamMetadata(imageWriteParam);
        }
        else {
            gifWritableStreamMetadata = (GIFWritableStreamMetadata)this.convertStreamMetadata(iioMetadata, imageWriteParam);
        }
        this.write(true, true, gifWritableStreamMetadata, iioImage, imageWriteParam);
        return;
    }
    
    @Override
    public void writeToSequence(IIOImage iioImage, ImageWriteParam imageWriteParam) throws IOException {
        if (this.stream == null) {
            throw new IllegalStateException("output == null!");
        }
        if (iioImage == null) {
            throw new IllegalArgumentException("image == null!");
        }
        if (iioImage.hasRaster()) {
            throw new UnsupportedOperationException("canWriteRasters() == false!");
        }
        if (!this.isWritingSequence) {
            throw new IllegalStateException("prepareWriteSequence() was not invoked!");
        }
        this.write(!this.wroteSequenceHeader, false, this.theStreamMetadata, iioImage, imageWriteParam);
        if (!this.wroteSequenceHeader) {
            this.wroteSequenceHeader = true;
        }
        ++this.imageIndex;
        return;
    }
    
    private boolean needToCreateIndex(RenderedImage renderedImage) {
        SampleModel sampleModel = renderedImage.getSampleModel();
        ColorModel colorModel = renderedImage.getColorModel();
        return sampleModel.getNumBands() != 1 || sampleModel.getSampleSize()[0] > 8 || colorModel.getComponentSize()[0] > 8;
    }
    
    private void write(boolean b, boolean b2, IIOMetadata iioMetadata, IIOImage iioImage, ImageWriteParam imageWriteParam) throws IOException {
        this.clearAbortRequest();
        RenderedImage renderedImage = iioImage.getRenderedImage();
        if (this.needToCreateIndex(renderedImage)) {
            renderedImage = PaletteBuilder.createIndexedImage(renderedImage);
            iioImage.setRenderedImage(renderedImage);
        }
        ColorModel colorModel = renderedImage.getColorModel();
        SampleModel sampleModel = renderedImage.getSampleModel();
        Rectangle o = new Rectangle(renderedImage.getMinX(), renderedImage.getMinY(), renderedImage.getWidth(), renderedImage.getHeight());
        Dimension o2 = new Dimension();
        computeRegions(o, o2, imageWriteParam);
        Object o3 = null;
        if (iioImage.getMetadata() != null) {
            o3 = new GIFWritableImageMetadata();
            this.convertMetadata("javax_imageio_gif_image_1.0", iioImage.getMetadata(), o3);
            if (o3.localColorTable == null) {
                o3.localColorTable = createColorTable(colorModel, sampleModel);
                if ((colorModel instanceof IndexColorModel)) {
                    int transparentPixel = ((IndexColorModel)colorModel).getTransparentPixel();
                    o3.transparentColorFlag = (transparentPixel != -1);
                    if (o3.transparentColorFlag) {
                        o3.transparentColorIndex = transparentPixel;
                    }
                }
            }
        }
        byte[] array;
        if (b) {
            if (iioMetadata == null) {
                throw new IllegalArgumentException("Cannot write null header!");
            }
            GIFWritableStreamMetadata gifWritableStreamMetadata = (GIFWritableStreamMetadata)iioMetadata;
            if (gifWritableStreamMetadata.version == null) {
                gifWritableStreamMetadata.version = "89a";
            }
            if (gifWritableStreamMetadata.logicalScreenWidth == -1) {
                gifWritableStreamMetadata.logicalScreenWidth = o2.width;
            }
            if (gifWritableStreamMetadata.logicalScreenHeight == -1) {
                gifWritableStreamMetadata.logicalScreenHeight = o2.height;
            }
            if (gifWritableStreamMetadata.colorResolution == -1) {
                gifWritableStreamMetadata.colorResolution = ((colorModel != null) ? colorModel.getComponentSize()[0] : sampleModel.getSampleSize()[0]);
            }
            if (gifWritableStreamMetadata.globalColorTable == null) {
                if (this.isWritingSequence && o3 != null && o3.localColorTable != null) {
                    gifWritableStreamMetadata.globalColorTable = o3.localColorTable;
                }
                else if (o3 == null || o3.localColorTable == null) {
                    gifWritableStreamMetadata.globalColorTable = createColorTable(colorModel, sampleModel);
                }
            }
            array = gifWritableStreamMetadata.globalColorTable;
            int n;
            if (array != null) {
                n = getNumBits(array.length / 3);
            }
            else if (o3 != null && o3.localColorTable != null) {
                n = getNumBits(o3.localColorTable.length / 3);
            }
            else {
                n = sampleModel.getSampleSize(0);
            }
            this.writeHeader(gifWritableStreamMetadata, n);
        }
        else if (this.isWritingSequence) {
            array = this.theStreamMetadata.globalColorTable;
        }
        else {
            throw new IllegalArgumentException("Must write header for single image!");
        }
        this.writeImage(iioImage.getRenderedImage(), o3, imageWriteParam, array, o, o2);
        if (b2) {
            this.writeTrailer();
        }
    }
    
    private void writeImage(RenderedImage image, GIFWritableImageMetadata gifWritableImageMetadata, ImageWriteParam imageWriteParam, byte[] a, Rectangle rectangle, Dimension dimension) throws IOException {
        image.getColorModel();
        SampleModel sampleModel = image.getSampleModel();
        int transparentColorFlag;
        if (gifWritableImageMetadata == null) {
            gifWritableImageMetadata = (GIFWritableImageMetadata)this.getDefaultImageMetadata(new ImageTypeSpecifier(image), imageWriteParam);
            transparentColorFlag = gifWritableImageMetadata.transparentColorFlag;
        }
        else {
            NodeList elementsByTagName = null;
            try {
                elementsByTagName = ((IIOMetadataNode)gifWritableImageMetadata.getAsTree("javax_imageio_gif_image_1.0")).getElementsByTagName("GraphicControlExtension");
            }
            catch (IllegalArgumentException ex) {}
            transparentColorFlag = (elementsByTagName != null && elementsByTagName.getLength() > 0);
            if (imageWriteParam != null && imageWriteParam.canWriteProgressive()) {
                if (imageWriteParam.getProgressiveMode() == 0) {
                    gifWritableImageMetadata.interlaceFlag = false;
                }
                else if (imageWriteParam.getProgressiveMode() == 1) {
                    gifWritableImageMetadata.interlaceFlag = true;
                }
            }
        }
        if (Arrays.equals(a, gifWritableImageMetadata.localColorTable)) {
            gifWritableImageMetadata.localColorTable = null;
        }
        gifWritableImageMetadata.imageWidth = dimension.width;
        gifWritableImageMetadata.imageHeight = dimension.height;
        if (transparentColorFlag != 0) {
            this.writeGraphicControlExtension(gifWritableImageMetadata);
        }
        this.writePlainTextExtension(gifWritableImageMetadata);
        this.writeApplicationExtension(gifWritableImageMetadata);
        this.writeCommentExtension(gifWritableImageMetadata);
        this.writeImageDescriptor(gifWritableImageMetadata, getNumBits((gifWritableImageMetadata.localColorTable == null) ? ((a == null) ? sampleModel.getSampleSize(0) : (a.length / 3)) : (gifWritableImageMetadata.localColorTable.length / 3)));
        this.writeRasterData(image, rectangle, dimension, imageWriteParam, gifWritableImageMetadata.interlaceFlag);
    }
    
    private void writeRows(RenderedImage renderedImage, LZWCompressor lzwCompressor, int x, int n, int y, int n2, int w, int i, int n3, int length, int n4, int n5, int n6) throws IOException {
        int[] iArray = new int[w];
        byte[] buf = new byte[length];
        Raster raster = (renderedImage.getNumXTiles() == 1 && renderedImage.getNumYTiles() == 1) ? renderedImage.getTile(0, 0) : renderedImage.getData();
        int j;
        int n7;
        for (; i < n4; ++n5, y += n2, i += n3) {
            if (n5 % n6 == 0) {
                if (this.abortRequested()) {
                    this.processWriteAborted();
                    return;
                }
                this.processImageProgress((float)n5 * 100.0f / (float)n4);
            }
            raster.getSamples(x, y, w, 1, 0, iArray);
            for (j = 0, n7 = 0; j < length; ++j, n7 += n) {
                buf[j] = (byte)iArray[n7];
            }
            lzwCompressor.compress(buf, 0, length);
        }
    }
    
    private void writeRowsOpt(byte[] buf, int offset, int n, LZWCompressor lzwCompressor, int i, int n2, int length, int n3, int n4, int n5) throws IOException {
        for (offset += i * n, n *= n2, ; i < n3; ++n4, offset += n, i += n2) {
            if (n4 % n5 == 0) {
                if (this.abortRequested()) {
                    this.processWriteAborted();
                    return;
                }
                this.processImageProgress((float)n4 * 100.0f / (float)n3);
            }
            lzwCompressor.compress(buf, offset, length);
        }
    }
    
    private void writeRasterData(RenderedImage renderedImage, Rectangle rectangle, Dimension dimension, ImageWriteParam imageWriteParam, boolean b) throws IOException {
        int x = rectangle.x;
        int y = rectangle.y;
        int width = rectangle.width;
        int height = rectangle.height;
        int width2 = dimension.width;
        int height2 = dimension.height;
        int sourceXSubsampling;
        int sourceYSubsampling;
        if (imageWriteParam == null) {
            sourceXSubsampling = 1;
            sourceYSubsampling = 1;
        }
        else {
            sourceXSubsampling = imageWriteParam.getSourceXSubsampling();
            sourceYSubsampling = imageWriteParam.getSourceYSubsampling();
        }
        SampleModel sampleModel = renderedImage.getSampleModel();
        int codeSize = sampleModel.getSampleSize()[0];
        if (codeSize == 1) {
            ++codeSize;
        }
        this.stream.write(codeSize);
        LZWCompressor o = new LZWCompressor(this.stream, codeSize, false);
        int n = sourceXSubsampling == 1 && sourceYSubsampling == 1 && renderedImage.getNumXTiles() == 1 && renderedImage.getNumYTiles() == 1 && sampleModel instanceof ComponentSampleModel && renderedImage.getTile(0, 0) instanceof ByteComponentRaster && renderedImage.getTile(0, 0).getDataBuffer() instanceof DataBufferByte;
        int n2 = 0;
        int max = Math.max(height2 / 20, 1);
        this.processImageStarted(this.imageIndex);
        if (b) {
            if (!(n == 0)) {
                ByteComponentRaster byteComponentRaster = (ByteComponentRaster)renderedImage.getTile(0, 0);
                byte[] data2 = ((DataBufferByte)byteComponentRaster.getDataBuffer()).getData();
                ComponentSampleModel componentSampleModel2 = (ComponentSampleModel)byteComponentRaster.getSampleModel();
                int n5 = componentSampleModel2.getOffset(x, y, 0) + byteComponentRaster.getDataOffset(0);
                int scanlineStride = componentSampleModel2.getScanlineStride();
                this.writeRowsOpt(data2, n5, scanlineStride, o, 0, 8, width2, height2, n2, max);
                if (this.abortRequested()) {
                    return;
                }
                int n6 = n2 + height2 / 8;
                this.writeRowsOpt(data2, n5, scanlineStride, o, 4, 8, width2, height2, n6, max);
                if (this.abortRequested()) {
                    return;
                }
                int n7 = n6 + (height2 - 4) / 8;
                this.writeRowsOpt(data2, n5, scanlineStride, o, 2, 4, width2, height2, n7, max);
                if (this.abortRequested()) {
                    return;
                }
                this.writeRowsOpt(data2, n5, scanlineStride, o, 1, 2, width2, height2, n7 + (height2 - 2) / 4, max);
            }
            else {
                this.writeRows(renderedImage, o, x, sourceXSubsampling, y, 8 * sourceYSubsampling, width, 0, 8, width2, height2, n2, max);
                if (this.abortRequested()) {
                    return;
                }
                int n3 = n2 + height2 / 8;
                this.writeRows(renderedImage, o, x, sourceXSubsampling, y + 4 * sourceYSubsampling, 8 * sourceYSubsampling, width, 4, 8, width2, height2, n3, max);
                if (this.abortRequested()) {
                    return;
                }
                int n4 = n3 + (height2 - 4) / 8;
                this.writeRows(renderedImage, o, x, sourceXSubsampling, y + 2 * sourceYSubsampling, 4 * sourceYSubsampling, width, 2, 4, width2, height2, n4, max);
                if (this.abortRequested()) {
                    return;
                }
                this.writeRows(renderedImage, o, x, sourceXSubsampling, y + sourceYSubsampling, 2 * sourceYSubsampling, width, 1, 2, width2, height2, n4 + (height2 - 2) / 4, max);
            }
        }
        else if (!(n == 0)) {
            Raster tile = renderedImage.getTile(0, 0);
            byte[] data = ((DataBufferByte)tile.getDataBuffer()).getData();
            ComponentSampleModel componentSampleModel = (ComponentSampleModel)tile.getSampleModel();
            this.writeRowsOpt(data, componentSampleModel.getOffset(x, y, 0), componentSampleModel.getScanlineStride(), o, 0, 1, width2, height2, n2, max);
        }
        else {
            this.writeRows(renderedImage, o, x, sourceXSubsampling, y, sourceYSubsampling, width, 0, 1, width2, height2, n2, max);
        }
        if (this.abortRequested()) {
            return;
        }
        this.processImageProgress(100.0f);
        o.flush();
        this.stream.write(0);
        this.processImageComplete();
        return;
    }
    
    private void writeHeader(String str, int n, int n2, int n3, int n4, int n5, boolean b, int n6, byte[] array) throws IOException {
        try {
            this.stream.writeBytes("GIF" + str);
            this.stream.writeShort((int)(short)n);
            this.stream.writeShort((int)(short)n2);
            int n7 = ((array != null) ? 128 : 0) | (n3 - 1 & 0x7) << 4;
            if (b) {
                n7 |= 0x8;
            }
            this.stream.write(n7 | n6 - 1);
            this.stream.write(n5);
            this.stream.write(n4);
            if (array != null) {
                this.stream.write(array);
            }
        }
        catch (IOException cause) {
            throw new IIOException("I/O error writing header!", cause);
        }
    }
    
    private void writeHeader(IIOMetadata iioMetadata, int n) throws IOException {
        GIFWritableStreamMetadata gifWritableStreamMetadata;
        if ((iioMetadata instanceof GIFWritableStreamMetadata)) {
            gifWritableStreamMetadata = (GIFWritableStreamMetadata)iioMetadata;
        }
        else {
            gifWritableStreamMetadata = new GIFWritableStreamMetadata();
            gifWritableStreamMetadata.setFromTree("javax_imageio_gif_stream_1.0", iioMetadata.getAsTree("javax_imageio_gif_stream_1.0"));
        }
        this.writeHeader(gifWritableStreamMetadata.version, gifWritableStreamMetadata.logicalScreenWidth, gifWritableStreamMetadata.logicalScreenHeight, gifWritableStreamMetadata.colorResolution, gifWritableStreamMetadata.pixelAspectRatio, gifWritableStreamMetadata.backgroundColorIndex, gifWritableStreamMetadata.sortFlag, n, gifWritableStreamMetadata.globalColorTable);
    }
    
    private void writeGraphicControlExtension(int n, boolean b, boolean b2, int n2, int n3) throws IOException {
        try {
            this.stream.write(33);
            this.stream.write(249);
            this.stream.write(4);
            int n4 = (n & 0x3) << 2;
            if (b) {
                n4 |= 0x2;
            }
            if (b2) {
                n4 |= 0x1;
            }
            this.stream.write(n4);
            this.stream.writeShort((int)(short)n2);
            this.stream.write(n3);
            this.stream.write(0);
        }
        catch (IOException cause) {
            throw new IIOException("I/O error writing Graphic Control Extension!", cause);
        }
    }
    
    private void writeGraphicControlExtension(GIFWritableImageMetadata gifWritableImageMetadata) throws IOException {
        this.writeGraphicControlExtension(gifWritableImageMetadata.disposalMethod, gifWritableImageMetadata.userInputFlag, gifWritableImageMetadata.transparentColorFlag, gifWritableImageMetadata.delayTime, gifWritableImageMetadata.transparentColorIndex);
    }
    
    private void writeBlocks(byte[] array) throws IOException {
        if (array != null && array.length > 0) {
            int min;
            for (int i = 0; i < array.length; i += min) {
                min = Math.min(array.length - i, 255);
                this.stream.write(min);
                this.stream.write(array, i, min);
            }
        }
    }
    
    private void writePlainTextExtension(GIFWritableImageMetadata gifWritableImageMetadata) throws IOException {
        if (gifWritableImageMetadata.hasPlainTextExtension) {
            try {
                this.stream.write(33);
                this.stream.write(1);
                this.stream.write(12);
                this.stream.writeShort(gifWritableImageMetadata.textGridLeft);
                this.stream.writeShort(gifWritableImageMetadata.textGridTop);
                this.stream.writeShort(gifWritableImageMetadata.textGridWidth);
                this.stream.writeShort(gifWritableImageMetadata.textGridHeight);
                this.stream.write(gifWritableImageMetadata.characterCellWidth);
                this.stream.write(gifWritableImageMetadata.characterCellHeight);
                this.stream.write(gifWritableImageMetadata.textForegroundColor);
                this.stream.write(gifWritableImageMetadata.textBackgroundColor);
                this.writeBlocks(gifWritableImageMetadata.text);
                this.stream.write(0);
            }
            catch (IOException cause) {
                throw new IIOException("I/O error writing Plain Text Extension!", cause);
            }
        }
    }
    
    private void writeApplicationExtension(GIFWritableImageMetadata gifWritableImageMetadata) throws IOException {
        if (gifWritableImageMetadata.applicationIDs != null) {
            Iterator<Object> iterator = gifWritableImageMetadata.applicationIDs.iterator();
            Iterator<Object> iterator2 = gifWritableImageMetadata.authenticationCodes.iterator();
            Iterator<Object> iterator3 = gifWritableImageMetadata.applicationData.iterator();
            while (iterator.hasNext()) {
                try {
                    this.stream.write(33);
                    this.stream.write(255);
                    this.stream.write(11);
                    this.stream.write((byte[])(byte[])iterator.next(), 0, 8);
                    this.stream.write((byte[])(byte[])iterator2.next(), 0, 3);
                    this.writeBlocks((byte[])(byte[])iterator3.next());
                    this.stream.write(0);
                }
                catch (IOException cause) {
                    throw new IIOException("I/O error writing Application Extension!", cause);
                }
            }
        }
    }
    
    private void writeCommentExtension(GIFWritableImageMetadata gifWritableImageMetadata) throws IOException {
        if (gifWritableImageMetadata.comments != null) {
            try {
                Iterator<Object> iterator = gifWritableImageMetadata.comments.iterator();
                while (iterator.hasNext()) {
                    this.stream.write(33);
                    this.stream.write(254);
                    this.writeBlocks((byte[])(byte[])iterator.next());
                    this.stream.write(0);
                }
            }
            catch (IOException cause) {
                throw new IIOException("I/O error writing Comment Extension!", cause);
            }
        }
    }
    
    private void writeImageDescriptor(int n, int n2, int n3, int n4, boolean b, boolean b2, int n5, byte[] array) throws IOException {
        try {
            this.stream.write(44);
            this.stream.writeShort((int)(short)n);
            this.stream.writeShort((int)(short)n2);
            this.stream.writeShort((int)(short)n3);
            this.stream.writeShort((int)(short)n4);
            int n6 = (array != null) ? 128 : 0;
            if (b) {
                n6 |= 0x40;
            }
            if (b2) {
                n6 |= 0x8;
            }
            this.stream.write(n6 | n5 - 1);
            if (array != null) {
                this.stream.write(array);
            }
        }
        catch (IOException cause) {
            throw new IIOException("I/O error writing Image Descriptor!", cause);
        }
    }
    
    private void writeImageDescriptor(GIFWritableImageMetadata gifWritableImageMetadata, int n) throws IOException {
        this.writeImageDescriptor(gifWritableImageMetadata.imageLeftPosition, gifWritableImageMetadata.imageTopPosition, gifWritableImageMetadata.imageWidth, gifWritableImageMetadata.imageHeight, gifWritableImageMetadata.interlaceFlag, gifWritableImageMetadata.sortFlag, n, gifWritableImageMetadata.localColorTable);
    }
    
    private void writeTrailer() throws IOException {
        this.stream.write(59);
    }
}
